/*
  amqnotifyfs: FUSE based binding filesystem sending events to ActiveMQ
  Copyright (C) 2008  Serge Emond <greyl at greyworld dot net>

  This program can be distributed under the terms of the GNU GPL.
  See the file COPYING.

  Based on FUSE 2.7.2's fusexmp_fh.c, original notice:
  ---
  FUSE: Filesystem in Userspace
  Copyright (C) 2001-2007  Miklos Szeredi <miklos@szeredi.hu>
  ---

  Tab width: 2; jed/utf-8

*/

/* #define FUSE_USE_VERSION 26 */

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#define _GNU_SOURCE

#include <fuse.h>
#include <ulockmgr.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <sys/time.h>
#ifdef HAVE_SETXATTR
#	include <sys/xattr.h>
#endif

#include <stddef.h>
#include <stdlib.h>
#include <assert.h>

#include "amqbridge.h"

enum AmqDestType {
	AMQ_DT_TOPIC,
		AMQ_DT_QUEUE
};

static struct settings {
	const char *progname;								 /* Name used to call me */
	const char *mntsrc;									 /* Source path */
	const char *mntdest;								 /* Destination path ("mount point") */
	
	int mntsrc_fd;											 /* FD pointing to source */
	
	/* ActiveMQ Settings */
	const char *amq_brokerUri;					 /* ActiveMQ connection string */
	enum AmqDestType amq_destType;			 /* Topic or queue */
	const char *amq_destUri;						 /* Topic/queue string */
	
} settings; /* static struct settings */


static const char *process_path(const char *path)
{
	if (path == NULL) /* possible? */
		return NULL;
	
	while (*path == '/')
		++path;
	
	if (*path == '\0')
		return ".";
	else
		return path;
} /* process_path */


static void *amqn_init()
{
	//int *ii;
	void * ud;
	
	assert(settings.mntsrc_fd > 0);
	
	if (fchdir(settings.mntsrc_fd) != 0) {
		fprintf(
            stderr,
            "Could not change working directory to '%s': %s\n",
            settings.mntsrc,
            strerror(errno)
            );
		fuse_exit(fuse_get_context()->fuse);
	}
	
	ud = amqinit(settings.amq_brokerUri,
							 settings.amq_destUri,
							 settings.amq_destType == AMQ_DT_TOPIC, 0);
	
	if (!ud) {
		fprintf(
            stderr,
            "ActiveMQ-cpp initialization failed, exiting\n");
		fuse_exit(fuse_get_context()->fuse);
	}
	
	/** Return pointer to "user_data" (e.g. fuse_get_context()->private_data */
	return ud;
}

static void amqn_destroy(void *private_data)
{
	if (private_data)
		amqexit(private_data);
}

static int amqn_getattr(const char *path, struct stat *stbuf)
{
	int res;
	
	path = process_path(path);
	
	res = lstat(path, stbuf);
	if (res == -1)
		return -errno;

	return 0;
}

static int amqn_fgetattr(const char *path, struct stat *stbuf,
												 struct fuse_file_info *fi)
{
	int res;

	(void) path;

	res = fstat(fi->fh, stbuf);
	if (res == -1)
		return -errno;

	return 0;
}

static int amqn_access(const char *path, int mask)
{
	int res;

	path = process_path(path);
	
	res = access(path, mask);
	if (res == -1)
		return -errno;

	return 0;
}

static int amqn_readlink(const char *path, char *buf, size_t size)
{
	int res;

	path = process_path(path);
	
	res = readlink(path, buf, size - 1);
	if (res == -1)
		return -errno;

	buf[res] = '\0';
	return 0;
}

static int amqn_opendir(const char *path, struct fuse_file_info *fi)
{
	DIR *dp;
	
	path = process_path(path);
	
	dp = opendir(path);
	if (dp == NULL)
		return -errno;

	fi->fh = (unsigned long) dp;
	return 0;
}

static inline DIR *get_dirp(struct fuse_file_info *fi)
{
	return (DIR *) (uintptr_t) fi->fh;
}

static int amqn_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
												off_t offset, struct fuse_file_info *fi)
{
	DIR *dp = get_dirp(fi);
	struct dirent *de;

	(void) path;
	seekdir(dp, offset);
	while ((de = readdir(dp)) != NULL) {
		struct stat st;
		memset(&st, 0, sizeof(st));
		st.st_ino = de->d_ino;
		st.st_mode = de->d_type << 12;
		if (filler(buf, de->d_name, &st, telldir(dp)))
			break;
	}

	return 0;
}

static int amqn_releasedir(const char *path, struct fuse_file_info *fi)
{
	DIR *dp = get_dirp(fi);
	(void) path;
	closedir(dp);
	return 0;
}

static int amqn_mknod(const char *path, mode_t mode, dev_t rdev)
{
	int res;

	path = process_path(path);
	
	if (S_ISFIFO(mode))
		res = mkfifo(path, mode);
	else
		res = mknod(path, mode, rdev);
	if (res == -1)
		return -errno;

	amq_msg_new(fuse_get_context()->private_data, path);
	
	return 0;
}

static int amqn_mkdir(const char *path, mode_t mode)
{
	int res;

	path = process_path(path);
	
	res = mkdir(path, mode);
	if (res == -1)
		return -errno;

	amq_msg_new(fuse_get_context()->private_data, path);
	
	return 0;
}

static int amqn_unlink(const char *path)
{
	int res;

	path = process_path(path);
	
	res = unlink(path);
	if (res == -1)
		return -errno;

	amq_msg_rem(fuse_get_context()->private_data, path);
	
	return 0;
}

static int amqn_rmdir(const char *path)
{
	int res;

	path = process_path(path);
	
	res = rmdir(path);
	if (res == -1)
		return -errno;

	amq_msg_rem(fuse_get_context()->private_data, path);
	
	return 0;
}

static int amqn_symlink(const char *from, const char *to)
{
	int res;

	from = process_path(from);
	to = process_path(to);
	
	res = symlink(from, to);
	if (res == -1)
		return -errno;

	amq_msg_new(fuse_get_context()->private_data, to);
	
	return 0;
}

static int amqn_rename(const char *from, const char *to)
{
	int res;

	from = process_path(from);
	to = process_path(to);
	
	res = rename(from, to);
	if (res == -1)
		return -errno;

	amq_msg_rem(fuse_get_context()->private_data, from);
	amq_msg_new(fuse_get_context()->private_data, to);
	
	return 0;
}

static int amqn_link(const char *from, const char *to)
{
	int res;

	from = process_path(from);
	to = process_path(to);
	
	res = link(from, to);
	if (res == -1)
		return -errno;

	amq_msg_new(fuse_get_context()->private_data, to);
	
	return 0;
}

static int amqn_chmod(const char *path, mode_t mode)
{
	int res;

	path = process_path(path);
	
	res = chmod(path, mode);
	if (res == -1)
		return -errno;

	amq_msg_new(fuse_get_context()->private_data, path);
	
	return 0;
}

static int amqn_chown(const char *path, uid_t uid, gid_t gid)
{
	int res;

	path = process_path(path);
	
	res = lchown(path, uid, gid);
	if (res == -1)
		return -errno;

	amq_msg_new(fuse_get_context()->private_data, path);
	
	return 0;
}

static int amqn_truncate(const char *path, off_t size)
{
	int res;

	path = process_path(path);
	
	res = truncate(path, size);
	if (res == -1)
		return -errno;

	amq_msg_new(fuse_get_context()->private_data, path);
	
	return 0;
}

static int amqn_ftruncate(const char *path, off_t size,
													struct fuse_file_info *fi)
{
	int res;

	(void) path;

	res = ftruncate(fi->fh, size);
	if (res == -1)
		return -errno;

	return 0;
}

static int amqn_utimens(const char *path, const struct timespec ts[2])
{
	int res;
	struct timeval tv[2];

	tv[0].tv_sec = ts[0].tv_sec;
	tv[0].tv_usec = ts[0].tv_nsec / 1000;
	tv[1].tv_sec = ts[1].tv_sec;
	tv[1].tv_usec = ts[1].tv_nsec / 1000;

	path = process_path(path);
	
	res = utimes(path, tv);
	if (res == -1)
		return -errno;

	return 0;
}

static int amqn_create(const char *path, mode_t mode, struct fuse_file_info *fi)
{
	int fd;

	path = process_path(path);
	
	fd = open(path, fi->flags, mode);
	if (fd == -1)
		return -errno;

	fi->fh = fd;
	return 0;
}

static int amqn_open(const char *path, struct fuse_file_info *fi)
{
	int fd;

	path = process_path(path);

	fd = open(path, fi->flags);
	if (fd == -1)
		return -errno;

	fi->fh = fd;
	return 0;
}

static int amqn_read(const char *path, char *buf, size_t size, off_t offset,
										 struct fuse_file_info *fi)
{
	int res;

	(void) path;
	res = pread(fi->fh, buf, size, offset);
	if (res == -1)
		res = -errno;

	return res;
}

static int amqn_write(const char *path, const char *buf, size_t size,
											off_t offset, struct fuse_file_info *fi)
{
	int res;

	(void) path;
	res = pwrite(fi->fh, buf, size, offset);
	if (res == -1)
		res = -errno;

	return res;
}

static int amqn_statfs(const char *path, struct statvfs *stbuf)
{
	int res;

	path = process_path(path);
	
	res = statvfs(path, stbuf);
	if (res == -1)
		return -errno;

	return 0;
}

static int amqn_flush(const char *path, struct fuse_file_info *fi)
{
	int res;

	(void) path;
	/* This is called from every close on an open file, so call the
	   close on the underlying filesystem.	But since flush may be
	   called multiple times for an open file, this must not really
	   close the file.  This is important if used on a network
	   filesystem like NFS which flush the data/metadata on close() */
	res = close(dup(fi->fh));
	if (res == -1)
		return -errno;

	return 0;
}

static int amqn_release(const char *path, struct fuse_file_info *fi)
{
	//(void) path;
	close(fi->fh);

	path = process_path(path);
	
	if (fi->flags & (O_WRONLY|O_RDWR)) {
		// Was probably openned writable
		//amqx();
		amq_msg_new(fuse_get_context()->private_data, path);
	}
	
	return 0;
}

static int amqn_fsync(const char *path, int isdatasync,
											struct fuse_file_info *fi)
{
	int res;
	(void) path;

#ifndef HAVE_FDATASYNC
	(void) isdatasync;
#else
	if (isdatasync)
		res = fdatasync(fi->fh);
	else
#endif
		res = fsync(fi->fh);
	if (res == -1)
		return -errno;

	return 0;
}

#ifdef HAVE_SETXATTR
/* xattr operations are optional and can safely be left unimplemented */
static int amqn_setxattr(const char *path, const char *name, const char *value,
												 size_t size, int flags)
{
	int res;
	
	path = process_path(path);
	
	res = lsetxattr(path, name, value, size, flags);
	if (res == -1)
		return -errno;
	return 0;
}

static int amqn_getxattr(const char *path, const char *name, char *value,
												 size_t size)
{
	int res;
	
	path = process_path(path);
	
	res = lgetxattr(path, name, value, size);
	if (res == -1)
		return -errno;
	return res;
}

static int amqn_listxattr(const char *path, char *list, size_t size)
{
	int res;
	
	path = process_path(path);
	
	res = llistxattr(path, list, size);
	if (res == -1)
		return -errno;
	return res;
}

static int amqn_removexattr(const char *path, const char *name)
{
	int res;
	
	path = process_path(path);
	
	res = lremovexattr(path, name);
	if (res == -1)
		return -errno;
	return 0;
}
#endif /* HAVE_SETXATTR */

static int amqn_lock(const char *path, struct fuse_file_info *fi, int cmd,
										 struct flock *lock)
{
	(void) path;

	return ulockmgr_op(fi->fh, cmd, lock, &fi->lock_owner,
										 sizeof(fi->lock_owner));
}

/** Options Stuff **/

enum OptionKey {
	OPTKEY_NONOPTION = -2,
		OPTKEY_UNKNOWN = -1,
		OPTKEY_HELP,
		OPTKEY_VERSION
};

static void dump_usage()
{
	printf("Usage: %s --broker-uri=URI --queue=NAME dir mountpoint\n"
				 "    OR %s --broker-uri=URI --topic=NAME dir mountpoint\n"
				 "    OR %s [options]\n"
				 "Options:\n"
				 "  -h, --help         Show help\n"
				 "  -V, --version      Show version\n"
				 "\n"
				 "  --broker-uri=URI   ActiveMQ broker's URI\n"
				 "\n"
				 "  --queue=NAME       Send notifications to queue NAME\n"
				 "  --topic=NAME       Send notifications to topic NAME\n"
				 "",
				 settings.progname,
				 settings.progname,
				 settings.progname);
} /* dump_usage() */

static int process_option(void *data, const char *arg, int key, struct fuse_args *outargs)
{
	switch ((enum OptionKey)key) {
	 case OPTKEY_HELP:
		dump_usage();
		exit(0);
		
	 case OPTKEY_VERSION:
		printf("%s\n", PACKAGE_STRING);
		exit(0);
		
	 case OPTKEY_NONOPTION:
		if (!settings.mntsrc) {
			printf("mntsrc: %s\n", arg);
			settings.mntsrc = arg;
			return 0;
		} else if (!settings.mntdest) {
			printf("mntdst: %s\n", arg);
			settings.mntdest = arg;
			return 1; /* leave this argument for fuse_main */
		} else {
			fprintf(stderr, "Too many arguments given\n");
			return -1;
		}
		
	 default:
		return 1;
	}
	
} /* process_option() */


/** End of Options Stuff **/

static struct fuse_operations amqn_oper = {
	.init     = amqn_init,
		.destroy  = amqn_destroy,
		.getattr	= amqn_getattr,
		.fgetattr	= amqn_fgetattr,
		.access		= amqn_access,
		.readlink	= amqn_readlink,
		.opendir	= amqn_opendir,
		.readdir	= amqn_readdir,
		.releasedir	= amqn_releasedir,
		.mknod		= amqn_mknod,
		.mkdir		= amqn_mkdir,
		.symlink	= amqn_symlink,
		.unlink		= amqn_unlink,
		.rmdir		= amqn_rmdir,
		.rename		= amqn_rename,
		.link		= amqn_link,
		.chmod		= amqn_chmod,
		.chown		= amqn_chown,
		.truncate	= amqn_truncate,
		.ftruncate	= amqn_ftruncate,
		.utimens	= amqn_utimens,
		.create		= amqn_create,
		.open		= amqn_open,
		.read		= amqn_read,
		.write		= amqn_write,
		.statfs		= amqn_statfs,
		.flush		= amqn_flush,
		.release	= amqn_release,
		.fsync		= amqn_fsync,
#ifdef HAVE_SETXATTR
		.setxattr	= amqn_setxattr,
		.getxattr	= amqn_getxattr,
		.listxattr	= amqn_listxattr,
		.removexattr	= amqn_removexattr,
#endif
		.lock		= amqn_lock,
};

int main(int argc, char *argv[])
{
	
	struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
	
	/* Fuse's option parser will store things here. */
	static struct OptionData {
		char *amq_brokerUri;
		char *amq_queue;
		char *amq_topic;
	} od = {NULL, NULL, NULL};
	
#define OPT2(one, two, key) \
	{one, -1, key}, \
	{two, -1, key}
#define OPT3(one, two, three, key) \
	{one, -1, key}, \
	{two, -1, key}, \
	{three, -1, key}
#define OPT_OFFSET2(one, two, offset, key) \
	{one, offsetof(struct OptionData, offset), key}, \
	{two, offsetof(struct OptionData, offset), key}
#define OPT_OFFSET3(one, two, three, offset, key) \
	{one, offsetof(struct OptionData, offset), key}, \
	{two, offsetof(struct OptionData, offset), key}, \
	{three, offsetof(struct OptionData, offset), key}
/*
 #define OPT_OFFSET4(one, two, three, four, offset, key) \
	{one, offsetof(struct OptionData, offset), key}, \
	{two, offsetof(struct OptionData, offset), key}, \
	{three, offsetof(struct OptionData, offset), key}, \
	{four, offsetof(struct OptionData, offset), key}
*/
	
	static const struct fuse_opt options[] = {
		OPT2("-h", "--help", OPTKEY_HELP),
		OPT2("-V", "--version", OPTKEY_VERSION),
		OPT_OFFSET2("--broker-uri=%s", "broker-uri=%s", amq_brokerUri, -1),
		OPT_OFFSET2("--queue=%s", "queue=%s", amq_queue, -1),
		OPT_OFFSET2("--topic=%s", "topic=%s", amq_topic, -1),
		FUSE_OPT_END
	};
	
	
	int fuse_main_return;
	
	settings.progname = argv[0];
	settings.mntsrc = NULL;
	settings.mntdest = NULL;
	
	/* Parse options */
	if (fuse_opt_parse(&args, &od, options, &process_option) == -1)
		return 1;
	
	/* Check that we have both a source directory and a moint point */
	if (!settings.mntsrc || !settings.mntdest) {
		fprintf(stderr, "Error: missing argument(s)\n");
		/* dump_usage(); */
		return 1;
	}
	
	if (od.amq_brokerUri) {
		settings.amq_brokerUri = od.amq_brokerUri;
	} else {
		fprintf(stderr, "Error: missing argument - broker URI is mandatory\n");
		/* dump_usage(); */
		return 1;
	}
	
	if (od.amq_topic && !od.amq_queue) {
		settings.amq_destType = AMQ_DT_TOPIC;
		settings.amq_destUri = od.amq_topic;
	} else if (!od.amq_topic && od.amq_queue) {
		settings.amq_destType = AMQ_DT_QUEUE;
		settings.amq_destUri = od.amq_queue;
	} else if (!od.amq_topic && !od.amq_queue) {
		fprintf(stderr, "Error: can't have both a topic and a queue\n");
		/* dump_usage(); */
		return 1;
	} else {
		fprintf(stderr, "Error: missing argument - a topic or a queue is required\n");
		/* dump_usage(); */
		return 1;
	}
	
	/* We want the kernel to do our access checks for us based on what getattr gives it. */
	fuse_opt_add_arg(&args, "-odefault_permissions");
	
	/* Open mount source for chrooting in amqn_init */
	settings.mntsrc_fd = open(settings.mntsrc, O_RDONLY);
	if (settings.mntsrc_fd == -1) {
		fprintf(stderr, "Could not open source directory\n");
		return 1;
	}
	
	/* Ignore the umask of the mounter on file creation */
	umask(0);
	fuse_main_return = fuse_main(args.argc, args.argv, &amqn_oper, NULL);
	
	fuse_opt_free_args(&args);
	close(settings.mntsrc_fd);
	
	return fuse_main_return;
}
