/*
 * amqnotifyfs: FUSE based binding filesystem sending events to ActiveMQ
 * Copyright (C) 2008  Serge Emond <greyl at greyworld dot net>
 *
 * This program can be distributed under the terms of the GNU GPL.
 * See the file COPYING.
 *
 * Original licence information of activemq-cpp's
 * SimpleProducer.cpp example follows
 *
 * ---
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ---
 */

#include <activemq/concurrent/Thread.h>
#include <activemq/concurrent/Runnable.h>
#include <activemq/concurrent/CountDownLatch.h>
#include <activemq/core/ActiveMQConnectionFactory.h>
#include <activemq/util/Integer.h>
#include <activemq/util/Config.h>
#include <activemq/util/Date.h>
#include <cms/Connection.h>
#include <cms/Session.h>
#include <cms/TextMessage.h>
#include <cms/BytesMessage.h>
#include <cms/MapMessage.h>
#include <cms/ExceptionListener.h>
#include <cms/MessageListener.h>
#include <stdlib.h>

using namespace activemq;
using namespace activemq::core;
using namespace activemq::util;
using namespace activemq::concurrent;
using namespace cms;
using namespace std;

////////////////////////////////////////////////////////////////////////////////
//class SimpleProducer : public Runnable {
class AmqBridge {
	private:

	 Connection* connection;
	 Session* session;
	 Destination* destination;
	 MessageProducer* producer;
	 bool useTopic;
	 bool clientAck;
	 std::string brokerURI;
	 std::string destURI;

	public:

	 AmqBridge( const std::string& brokerURI,
							const std::string& destURI,
							bool useTopic = false,
							bool clientAck = false ) {
		 connection = NULL;
		 session = NULL;
		 destination = NULL;
		 producer = NULL;
		 this->useTopic = useTopic;
		 this->brokerURI = brokerURI;
		 this->destURI = destURI;
		 this->clientAck = clientAck;
	 }

	 virtual ~AmqBridge() {
		 cleanup();
	 }
	 
	 virtual void connect() {
		 // Create a ConnectionFactory
		 ActiveMQConnectionFactory* connectionFactory =
			 new ActiveMQConnectionFactory( brokerURI );
		 
		 // Create a Connection
		 connection = connectionFactory->createConnection();
		 connection->start();
		 
		 // free the factory, we are done with it.
		 delete connectionFactory;
		 
		 // Create a Session
		 if( clientAck ) {
			 session = connection->createSession( Session::CLIENT_ACKNOWLEDGE );
		 }
		 else {
			 session = connection->createSession( Session::AUTO_ACKNOWLEDGE );
		 }
		 
		 // Create the destination (Topic or Queue)
		 if( useTopic ) {
			 destination = session->createTopic( destURI );
		 }
		 else {
			 destination = session->createQueue( destURI );
		 }
		 
		 // Create a MessageProducer from the Session to the Topic or Queue
		 producer = session->createProducer( destination );
		 producer->setDeliveryMode( DeliveryMode::NON_PERSISTENT );
		 
	 } // connect()
	 
	 virtual void postEvent(const std::string& text) {
		 TextMessage * msg = session->createTextMessage(text);
		 
		 producer->send(msg);
		 
		 delete msg;
	 } // postEvent()
	 
	private:

	 void cleanup() {

		 // Destroy resources.
		 try{
			 if( destination != NULL ) delete destination;
		 }catch ( CMSException& e ) { e.printStackTrace(); }
		 destination = NULL;

		 try{
			 if( producer != NULL ) delete producer;
		 }catch ( CMSException& e ) { e.printStackTrace(); }
		 producer = NULL;

		 // Close open resources.
		 try{
			 if( session != NULL ) session->close();
			 if( connection != NULL ) connection->close();
		 }catch ( CMSException& e ) { e.printStackTrace(); }

		 try{
			 if( session != NULL ) delete session;
		 }catch ( CMSException& e ) { e.printStackTrace(); }
		 session = NULL;

		 try{
			 if( connection != NULL ) delete connection;
		 }catch ( CMSException& e ) { e.printStackTrace(); }
		 connection = NULL;
	 }
};
